/*
Programmbeispiel fr serielle Ausgabe USART3 per ISR

Alle Funktionen ohne Gewhr
Michael Nowak
2020-10-24
www.mino-elektronik.de
*/

#define _USART3
#define PRIO_USART3     10          // mittlere Prioritt
#include "stm32h7xx.h"

#define BIT(x) (1<<x)
enum gpio_modus {IN, OUT, AF, ANALOG};
#define CLR_MODER       ANALOG      // Lschmaske, beide Bits gesetzt
enum gpio_speed {LOW_SPEED, MEDIUM_SPEED, HIGH_SPEED, V_HIGH_SPEED};
enum gpio_pullup {NO_PULL, PULL_UP, PULL_DOWN};

// PortB
#define TxD3_Pin        10
#define RxD3_Pin        11

#define AFR_USART3      7           // fr rx/tx lt. Datenblatt
#define APB1_TEILER     4           // 1/4 SystemCoreClock

#define DEF_BAUDRATE    115200
#define RX3_BUF_LEN  	  0x100
#define TX3_BUF_LEN	    0x100
#define RX3_BUF_MASK  	(RX3_BUF_LEN-1)
#define TX3_BUF_MASK	  (TX3_BUF_LEN-1)

static uint8_t usart3_rx_puffer[RX3_BUF_LEN];
static uint8_t usart3_tx_puffer[TX3_BUF_LEN];
static uint32_t usart3_rx_schreib_zeiger,
                usart3_rx_lese_zeiger;
static uint32_t usart3_tx_schreib_zeiger,
                usart3_tx_lese_zeiger;
static uint8_t ser_status;

char set_baudrate3(int baud)
{
  USART3->BRR = SystemCoreClock/APB1_TEILER/baud;
  return(0);
}
void USART3_IRQHandler(void)
{
uint32_t temp, sr_temp;
uint32_t temp_zeiger;
  sr_temp = USART3->ISR;
  if(sr_temp & USART_ISR_TXE_TXFNF) {         // Tx-Flag testen
    if(usart3_tx_schreib_zeiger != usart3_tx_lese_zeiger) { // noch zeichen im puffer
    USART3->TDR = usart3_tx_puffer[usart3_tx_lese_zeiger++];	// Zeichen ausgeben
      usart3_tx_lese_zeiger &= TX3_BUF_MASK;
    } else  {
      USART3->CR1 &= ~USART_CR1_TXEIE;        // sonst weitere Interrupts sperren */
      __ISB();                                // Ausfhrung abwarten
    }  
  }
  if(sr_temp & USART_ISR_RXNE_RXFNE) {        // Rx-Flag testen
    temp = USART3->RDR;                       // Zeichen lesen
    USART3->ICR = USART_ICR_ORECF;            // OVE-Bit immer loeschen
    temp_zeiger = (usart3_rx_schreib_zeiger + 1) & RX3_BUF_MASK;
    if( temp_zeiger != usart3_rx_lese_zeiger) {
      usart3_rx_puffer[usart3_rx_schreib_zeiger] = temp;
      usart3_rx_schreib_zeiger = temp_zeiger;
    }
  }
  if(sr_temp & USART_ISR_ORE) {               // Ueberlauf-Flag testen
    USART3->ICR = USART_ICR_ORECF;            // OVE-Bit immer loeschen
  }
}

void clr_tx3_buf()
{
  USART3->CR1 &= ~(USART_CR1_TXEIE);          // TXIE abschalten
  __ISB();
  usart3_tx_lese_zeiger = 0;
  usart3_tx_schreib_zeiger = 0;
}

void clr_rx3_buf()
{
  USART3->CR1 &= ~(USART_CR1_RXNEIE);         // RXIE abschalten
  __ISB();
  usart3_rx_lese_zeiger = 0;
  usart3_rx_schreib_zeiger = 0;
  USART3->CR1 |= (USART_CR1_RXNEIE);          // RXIE einschalten
}

void init_usart3(void)
{
uint32_t tmp;  
	RCC->AHB4ENR |= RCC_AHB4ENR_GPIOBEN;        // PortB aktivieren
  RCC->APB1LENR |= RCC_APB1LENR_USART3EN;
  __ISB();
// AF2 fr USART3    
  tmp = GPIOB->MODER & ~(CLR_MODER << (TxD3_Pin * 2) | CLR_MODER << (RxD3_Pin * 2));
  GPIOB->MODER = tmp | (AF << (TxD3_Pin * 2) | AF << (RxD3_Pin * 2));

// PB10+PB11 USART3 alternative Funktionen mit den IO-Pins verbinden
  GPIOB->AFR[1] |=  (AFR_USART3 << ((TxD3_Pin-8) * 4) | AFR_USART3 << ((RxD3_Pin-8) * 4));

// alle weiteren Einstellungen default 8N1 
  set_baudrate3(DEF_BAUDRATE);

// enable UART Interrupt-Vector
  NVIC_SetPriority(USART3_IRQn,PRIO_USART3);
  NVIC_EnableIRQ(USART3_IRQn);                // USART3 aktivieren
    
  clr_rx3_buf();
  clr_tx3_buf();
// RXIE und Rx + Tx einschalten  
  USART3->CR1 |= USART_CR1_RXNEIE | USART_CR1_UE | USART_CR1_RE | USART_CR1_TE;
  ser_status = 1;                             // USART3 ist initialisiert
}	

void write_tx3_buf(char c)
{
uint32_t temp_zeiger;
  if(!ser_status) init_usart3() ;
  do {
    temp_zeiger = (usart3_tx_schreib_zeiger + 1) & TX3_BUF_MASK;
  } while( temp_zeiger == usart3_tx_lese_zeiger); // warten, bis wieder platz ist
  usart3_tx_puffer[usart3_tx_schreib_zeiger] = c;
  usart3_tx_schreib_zeiger = temp_zeiger;
  USART3->CR1 |= USART_CR1_TXEIE;                 // per Interrupt ausgeben
  
}


char read_rx3_buf()
{
int32_t temp = 0;                                 // default bei leerem Puffer
  if(!ser_status) init_usart3() ;
  if(usart3_rx_lese_zeiger != usart3_rx_schreib_zeiger) {
    temp = usart3_rx_puffer[usart3_rx_lese_zeiger++];
    usart3_rx_lese_zeiger &= RX3_BUF_MASK;
  }
  return(temp);
}


uint32_t read_rx3_buf_count()
{
	return((usart3_rx_schreib_zeiger - usart3_rx_lese_zeiger) & RX3_BUF_MASK);
}	

int32_t putchar(int32_t x)
{
  if((uint8_t)x == 10) write_tx3_buf(13); 
  write_tx3_buf((uint8_t)x);
  return(x);
}

void sende_string(char *s)
{
  while(*s) write_tx3_buf(*(s++));
  putchar(10);
} 
